# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.16.3
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# + [markdown] jp-MarkdownHeadingCollapsed=true
# # Running eazypy on ray
#
# We will run `eazypy` with spherex referece catalog on the ray cluster (https://www.ray.io/).
# The original `eazypy` (https://github.com/gbrammer/eazy-py) supports multi-core execution uwing python's `multiprocessing` module (https://docs.python.org/3/library/multiprocessing.html). This makes tricky to run eazypy on the multi-node cluster.
# To make things simple, I have modified the eazypy (https://github.com/leejjoon/eazy-py) and replaced the use of multiprocessing with `joblib` module (https://joblib.readthedocs.io/en/stable/). `joblib` supports multiple backends including multiprocessing itself.
# Furthermore, `ray` does provide its own implementation of joblib backend. On the other hand, using joblib with ray backend likely here may not be efficitn enough as this does not fully utilize ray's shared memory feautures.
#
# -

# # Setting up ray cluster
#
# All the nodes should have ray installed. We will simply use python 3.10 but the version should not matter. We will create a conda environment for ray together with two additional packages that are required.
#
# ```sh
# conda create -n ray_node_py310 python=3.10 smart_open virtualenv
# conda activate ray_node_py310
# pip install ray[default]
# ```
#
# On the head node, we start the master ray instance by
#
# ```sh
# ray start --head --port 6379
# ```
#
# For the rest of the node,
#
# ```sh
# ray start --address='IP_ADDRESS_OF_THE_HEAD_NODE:6379'
# ```
#
# The head and sibling nodes need to communicate. And several ports need to be open. The list of ports that need to be open is a bit ambiguous for now. And to make things easy, we simply open all the ports between the nodes. 

# # Runtime evironment
#
# `ray` allows you to set up your own runtimeenvironment to run your tasks. Your runtime environment will include python virtual environet with list of packages to be installd and the contents of the working directory. To run our eazypy taks, we will set up a virtual enviroment with `ray`, `joblib`, `jax` and `jaxopt` and the customized `eazy` package and the `dust_attenumation` package from their github repo (both are wheel-packaged and uloaded to KASI's minio storage which is publicly available. The workdir content can be set up from a zip file, and the zip file is also uploaded to the minio server, which contains all the necessary files to run eazypy. 
#
# Note that the runtime environment won't include reference catalogs. Reference catalogs will be accessed by the node this notebook is running (see below) and will be transfered to the ray nodes by ray's own shared memory behind the scene.

# # Running this notebook
#
# To run this notebook, your environment require these packages. `ray`, `pyarrow`, `pandas` `joblib`, `astropy` and `eazypy` (the original eazypy should suffice).
#
# This notebook should be run on the machine which has an access to the reference catalog and all the output hdf files will be saved in the same node. The reference catalogs are under `catdir` and the output hdf files will be saved under `outdir`.
#
# The template will also be read from the notebook. For now, we will use a pre-build template for the band set of reference catalog.
#
# Note that, for now, the node you are running this notebook should be a member of ray cluster. If you want to run this notebook from a node that is not a member of the ray cluster, you need to install `ray[client]` package and `ray.init` with a pointer to the head node.
#

import ray


# +
# We will join the ray cluster with the runtime environment definition for eazypy.

uri_root = "https://archive.kasi.re.kr/kgmt/public/spherex_eazy/"

ray_env_requirements = ["ray", "joblib", "jax", "jaxopt"]

ray_env_requirements.extend([
    uri_root + "dust_attenuation-0.5.dev12+gdf42887-py3-none-any.whl",
    uri_root + "eazy-0.6.9.dev14+gd5c524c-py3-none-any.whl"
])

runtime_env = {"pip": ray_env_requirements,
               "working_dir": uri_root + "spherex_eazypy_data.zip"}

ray.init(runtime_env=runtime_env)

# -

from pprint import pprint
resources = ray.cluster_resources()
pprint(resources)


# +
# we will first run a simple test task so that runtime environement are set up in each node. Note that
# this is not necessary. We are just doing this to make sure that the runtime env is correctly set up.

@ray.remote
def check_env(nodename):
    import eazy
    from pathlib import Path
    files = Path(".").glob("*")
    return nodename, list(files)

refs = []
for node in ray.nodes():
    nodename = node["NodeName"]
    resources = {f'node:{nodename}': 0.01}
    ref = check_env.options(resources=resources).remote(nodename)
    refs.append(ref)
    
for nodename, files in ray.get(refs):
    print(nodename, ", ".join(sorted(p.name for p in files)))

# This may take a while since the packages need to be installed and workind dir contents need to be downloaded.
# This will print out contents of the working directory set up in each node.
# -

from pathlib import Path
import numpy as np
from astropy.table import Table


# +
# We will register ray backend for the joblib. Somehow the defulat RayBackend from ray does not support generator.
# We simply overwrite the class attribute so that generator can be supported.

import ray
from joblib import parallel_backend
from joblib.parallel import register_parallel_backend
from ray.util.joblib.ray_backend import RayBackend

class RayBackend2(RayBackend):
    supports_return_generator = True

# The registration itself should be done at each node and will be executed inside the ray task.


# +
# This is is the main task that will be run on ray. We will overwrite the nproc parameter later. Note
# that the function need to be wraped by `ray.remote` and this will be done later.

def do_single_catalog_simple(catfile, cattab, tempfilt_data, nproc=-1):

    import eazy

    register_parallel_backend("ray", RayBackend2)

    # %%

    # ## 2. Parameters for eazy
    # outputfile = 'SPHERExRefCat_eazypy_specz_photoz'
    params = {}
    params['CATALOG_FILE'] = cattab
    params['MAIN_OUTPUT_FILE'] = "" # outputfile
    params['FIX_ZSPEC'] = 'n'
    #params['TEMPLATES_FILE'] = './templates.spherex/brown_cosmos_ebv.param'
    params['TEMPLATES_FILE'] = './templates/fsps_full/tweak_fsps_QSF_12_v3.param'
    params['TEMPLATE_COMBOS'] = 1  # one template at a time [NOT WORKING], default = 99 or 'a'
    params['N_MIN_COLORS'] = 3     # EAZY default value : 5
    params['APPLY_PRIOR'] = 'y'
    params['PRIOR_FILE'] = 'templates/prior_K_TAO.dat'
    params['PRIOR_FILTER'] = 163   # 2MASS Ks band

    params['Z_MAX'] = 6.0
    params['Z_STEP'] = 0.01

    params['VERBOSITY'] = 0

    # %%

    ez = eazy.photoz.PhotoZ(
        param_file='zphot.param.default.RefCat.eazypy',
        translate_file='zphot.RefCat.translate',
        zeropoint_file=None,
        params=params,
        load_prior=True,
        load_products=False,
        tempfilt_data=tempfilt_data
        )

    with parallel_backend('ray'):
        ez.fit_catalog(n_proc=nproc, fitter="nnls_jax", do_tqdm=False)

    # FIXME: replace the CATALOG_FILE to a string

    return catfile, ez



# +
# This is to save the output as a hdf. This will be run locally.

import eazy.hdf5

def hdf_save(catfile, i, ez):

    print("saving", catfile, i)
    ez.param.params["CATALOG_FILE"] = catfile.name
    ext = f"_{i:03d}.hdf"
    outname = outdir / (catfile.with_suffix("").name + ext) 
    eazy.hdf5.write_hdf5(ez, h5file=outname)



# +
# We will setup the list of reference catalog that will be processed. Some reference catalog is too big and
# better to be split to reduce the memory pressure. If any of the file has number of rows larger than `maxrow`, we will split the catalog.
# We also set number of nproc here.

# The number of avaialbe cores in the cluster divided by `nproc` gives you a number of catalog that needs to be simultanesouly processed,
# which is `MAX_NUM_PENDING_CATALOGS`. `MAX_NUM_PENDING_CATALOGS` can be larger than the value defined here, but too large number will 
# increase the memory of the queue.

catdir = Path("./catdir")
outdir = Path("./outdir")

catfiles = list(sorted(catdir.glob("Gaia*.parquet")))

catfiles_to_run = catfiles[:10] # we will only run the first 10 files for a test.
# -

# # PARAMETERS for RAY TASKS
#
# In the cell below, you need to adjust the parameters so that they are suitable for your environment. `nproc` is a number of cores that will be used to process a single catalog, and `MAX_NUM_PENDING_CATALOGS` is a number of catalogs to process simultaneously. `nproc` x `MAX_NUM_PENDING_CATALOGS` should be similar to number of total cores in your ray cluster. A large catalog will be split into multiple smaller ones. `maxrow` sets the maximum number of rows for the split.

maxrow = 400_000 # 400k rows take about 3 minutes on 32 cores
nproc = 16
cpus = ray.cluster_resources()["CPU"]
MAX_NUM_PENDING_CATALOGS = int(cpus // nproc)
print("# of cores in the cluster:", cpus)
print("NPROC:", nproc)
print("MAX_NUM_PENDING_CATALOGS:", MAX_NUM_PENDING_CATALOGS)
print("N_catalogs:", len(catfiles_to_run))


# +
# This is the actual function that will be run on the ray cluster. Note that we override the value of npc.

@ray.remote
def do_single_catalog(catfile, cattab, tempfilt_data, nproc=nproc): # to override nproc
    return do_single_catalog_simple(catfile, cattab, tempfilt_data, nproc=nproc)



# -

# We load the pre-build template for the spherex reference catalog.
tempfilt_data = np.load("tempfilt.npy")
tempfilt_data_r = ray.put(tempfilt_data)


# +
# A generator to open the reference catalog on the flux and convert it to the catalog for eazypy.
# This will also split the large catalog into smaller ones.

from spherex_refcat_to_eazy import SPHERExRef2Eazy

def get_cattab(catfiles, maxrow=None):
    spherex2eazy = SPHERExRef2Eazy().get_eazy_table

    for I, catfile in enumerate(catfiles):
        phot = Table.read(catfile)

        if (maxrow is None) or len(phot) < maxrow:
            cattab = spherex2eazy(phot)
            yield I, catfile, 0, cattab
        else:
            count = 0
            for i, istart in enumerate(range(0, len(phot), maxrow)):
                cattab = spherex2eazy(phot[istart:istart+maxrow])
                yield I, catfile, i, cattab
                count += 1
                # if TEST_N is not None and count > TEST_N: return



# -

def get_batch_args(catfiles, maxrow): # generator fro function arguments
    for I, catfile, i, cattab in get_cattab(catfiles, maxrow=maxrow):
        yield (I, catfile, i), cattab, tempfilt_data_r


# +
# The generator below and the commented out contents in the next cell is to run the eazypy in a blocking way
# (the cell will run until all the processing is done). On the othe hand, you may prefer to us asyncio version below.
# Your jupyter session will be still interactive while eazypy is being run in the background. 

# A generator to push a catalog to the ray cluster with a maximum number pending catalogs.

def run_with_ntask(ray_func, batches, ntask):
    """run the task with arguments list of batches, with maximum simultaneously running task smaller than ntask"""
    
    result_refs = []

    for args in batches:
        # catfile, cattab = ray.get(ref)
        if len(result_refs) >= ntask:
            ready_refs, result_refs = ray.wait(result_refs, num_returns=1)
            for _ in ray.get(ready_refs):
                yield _

        task = ray_func.remote(*args)
        result_refs.append(task)

    while result_refs:
        ready_refs, result_refs = ray.wait(result_refs, num_returns=1)

        for _ in ray.get(ready_refs):
            yield _
            


# +
# This will run eazypy on the ray cluster, and it will take long.

# batche_args = get_batch_args(catfiles_to_run, maxrow=maxrow)
# for (I, catfile1, i1), ez in run_with_ntask(do_single_catalog, batche_args, MAX_NUM_PENDING_CATALOGS):
#     hdf_save(catfile1, i1, ez)


# +
# asyncio version

async def async_run_with_ntask(ray_func, batches, ntask):
    """run the task with arguments list of batches, with maximum simultaneously running task smaller than ntask"""
    
    result_refs = set()

    for args in batches:
        # catfile, cattab = ray.get(ref)
        if len(result_refs) >= ntask:
            ready_refs, result_refs = await asyncio.wait(result_refs)
            for _ in ready_refs:
                yield await _

        task = ray_func.remote(*args)
        result_refs.add(task)

    while result_refs:
        ready_refs, result_refs = await asyncio.wait(result_refs)

        for _ in ready_refs:
            yield await _
# -




# +
# easypy will be running as an asyncio background task.

import asyncio
taskmap = [False] * len(catfiles_to_run)
async def run_tasks():
    batch_args = get_batch_args(catfiles_to_run, maxrow=maxrow)
    global taskmap
    async for (I, catfile1, i1), ez in async_run_with_ntask(do_single_catalog, batch_args, MAX_NUM_PENDING_CATALOGS):
        hdf_save(catfile1, i1, ez)
        taskmap[I] = True

#task = await run_tasks()
import time
ctime = time.time()
task = asyncio.create_task(run_tasks())

# +
# Run this cell to check the status
# No that the catalog is marked to be processed if its first batch is done.

n_catalog = len(catfiles_to_run)
n_finished = np.sum(taskmap)
elapsed_hour = (time.time() - ctime) / 3600.
remaining_hour = elapsed_hour / n_finished * (n_catalog - n_finished)

print(f"{n_finished} out of {n_catalog} catalogs are processed for {elapsed_hour:.2f}h.")
print(f"{remaining_hour:.2f}h to process the rest.")

# -

import matplotlib.pyplot as plt
plt.plot(taskmap)
#print(taskmap)


# Note that this will block jupyter notebook until the processing is done. This is meand to do the garbage collecting 
# after the processing is done.
await task

ray.shutdown()

